#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"

#include "../port/sd.h"

extern SDifc sdmv88alp01ifc;

typedef struct Ctlr Ctlr;
typedef struct Reg Reg;

struct Ctlr
{
	Lock;
	Rendez vous;

	int intr;
	int irq;
	int tbdf;
	int rid;
	int enabled;
	SDev *sdev;
	Pcidev *pcidev;

	void *mmio;
};

struct Reg
{
	ushort	addrl;
	ushort	addrh;
	ushort	blocksz;
	ushort	blockc;
	ushort	argl;
	ushort	argh;
	ushort	transfer;
	ushort	command;
	ushort	response[8];
	ushort	buffer[2];
	ushort	state0;
	ushort	state1;
	ushort	host;
	ushort	block_gap;
	ushort	clock;
	ushort	timeout;
	ushort	intr_status;
	ushort	intr_err_status;
	ushort	intr_en;
	ushort	intr_err_en;
//	ushort	"normal interrupt status interrupt enable";
//	ushort	"error interrupt status enable"
//	ushort	"cmd12 error status";

//	0x40,	"capabilities 0",
//	0x42,	"capabilities 1",
//	0x44,	"capabilities 2",
//	0x46,	"capabilities 3",
//	0x48,	"current 0",
//	0x4a,	"current 1",
//	0x4c,	"current 2",
//	0x4e,	"current 3",

//	0x60,	"io control",

//	0x6a,	"command 1",
//	0xFC,	"slot interrupt",
//	0xFE,	"host control version",
};

enum {
	STins	= 0x0001,
	STstable	= 0x0002,
	STdet	= 0x0004,
	STwp	= 0x0008,
	STonline	= STins|STstable|STdet,
	CFcrc	= 0x08,
	CFindex	= 0x10,
	Irespond	= 1<<0,
	Ierror	= 1<<15,
};

/* sd command set */
enum {
	Creset = 0,
	Ccid = 2,
	Csrca = 3,
	Cswitch = 6,
	Cselect = 7,
	Ccond = 8,
	Cvolt = 11,
	Cstop = 12,
	Cinactive = 15,
	Capp = 55,
};

struct cmdinfo {
	uchar flags;
};
enum {
	FRnone = 0,
	FR136 = 1,
	FR48 = 2,
	FR48b = 3,
	FRmask = 3,
	Fdata = 4,
	Fcrc = 8,
	Fopcode = 16,
};

struct cmdinfo cmdinfo[] = {
[Creset]	FRnone,
[Ccid]	FR136|Fcrc,
[Csrca]	FR48|Fcrc|Fopcode,
[Ccond]	FR48|Fcrc|Fopcode,
[Cvolt]	FR48|Fcrc|Fopcode,
[Capp]	FR48|Fcrc|Fopcode,
};

static int
gotintr(void *a){
	Ctlr *ctlr;
	ctlr = a;
	return ctlr->intr;
}

ulong
mv88cmd(Ctlr *ctlr, uint cmd, ulong arg, uchar *response){
	Reg *reg;
	ushort cmdreg, intr, resp;
	int flags;
	int i;

	intr = 0;
	flags = cmdinfo[cmd].flags;
	cmdreg = (cmd & 0x7f) << 8;
	cmdreg |= flags & FRmask;
	if(flags & Fcrc)
		cmdreg |= CFcrc;
	if(flags & Fopcode)
		cmdreg |= CFindex;
	intr |= Irespond;

	reg = ctlr->mmio;

	reg->argl = arg & 0xffff;
	reg->argh = arg >> 16;
	reg->intr_status = ~0x100;
	reg->intr_err_status = 0xffff;
	reg->command = cmdreg;

	reg->intr_en = (reg->intr_en & 0x100) | intr;
	reg->intr_err_en = 0xffff;

	print("mv88alp01: cmd %d\n", cmd);

	if((flags & FRmask) == 0)
		return 1;

	ctlr->intr = 0;
//	reg->intr_en = 0xffff;
//	reg->intr_err_en = 0xffff;

	print("mv88alp01: cmd %d intr_en %04hux intr_err_en %04hux\n", cmd, reg->intr_en, reg->intr_err_en);
	sleep(&ctlr->vous, gotintr, ctlr);
	print("mv88alp01: awake, intr %hux err %hux\n", reg->intr_status, reg->intr_err_status);

	if(flags & FRmask && response != nil){
		for(i=0;i<8;i++){
			resp = reg->response[i];
			response[i*2] = resp & 0xff;
			response[i*2+1] = resp >> 8;
		}
	}
	return 1;
}

static void
mv88interrupt(Ureg*, void *a){
	Ctlr *ctlr;
	Reg *reg;

	ctlr = a;
	ilock(ctlr);

	reg = ctlr->mmio;
	iprint("intr hit %04hux %04hux\n", reg->intr_status, reg->intr_err_status);
	ctlr->intr = reg->intr_status << 16 | reg->intr_err_status;
	reg->intr_status = 0;
	reg->intr_err_status = 0;
	reg->intr_en = 0;
	reg->intr_err_en = 0;
	wakeup(&ctlr->vous);

	iunlock(ctlr);
}

static SDev*
mv88pnp(void)
{
	ulong io, *mem;
	Ctlr *ctlr;
	Pcidev *p;
	SDev *head, *tail, *sdev;
	static int ctrlno;

	p=nil;
	head=nil;
	tail=nil;
	while((p = pcimatch(p, 0x11ab, 0x4101)) != nil){
		print("Marvell 88ALP01: SD card interface found\n");
		if((sdev = malloc(sizeof(SDev))) == nil)
			continue;
		if((ctlr = malloc(sizeof(Ctlr))) == nil){
			free(sdev);
			continue;
		}
		memset(sdev, 0, sizeof(SDev));
		memset(ctlr, 0, sizeof(Ctlr));
		io = p->mem[0].bar & ~ 0x0f;
		mem = (ulong*)vmap(io, p->mem[0].size);
		if(mem == 0){
			print("sdmv88alp01: address 0x%lux in use\n", io);
			free(sdev);
			free(ctlr);
			continue;
		}
		ctlr->rid = p->rid;
		sdev->ifc = &sdmv88alp01ifc;
		sdev->ctlr = ctlr;
		sdev->nunit=1;
		sdev->idno='S';
		ctlr->sdev = sdev;
		ctlr->irq = p->intl;
		ctlr->tbdf = p->tbdf;
		ctlr->pcidev = p;
		ctlr->mmio = mem;
		ctlr->enabled = 0;
		ctrlno++;
		if(head)
			tail->next = sdev;
		else
			head = sdev;
		tail = sdev;
	}
	return head;
}

static int
mv88enable(SDev *sdev)
{
	char name[32];
	Ctlr *ctlr;
	Reg *reg;


	ctlr = sdev->ctlr;
	if(ctlr->enabled)
		return 1;
	reg = ctlr->mmio;
	reg->intr_en = 0;
	snprint(name, sizeof name, "%s (%s)", sdev->name, sdev->ifc->name);
	intrenable(ctlr->irq, mv88interrupt, ctlr, ctlr->tbdf, name);
	print("sd%c: enable %s\n", sdev->idno, name);
	ctlr->enabled = 1;
	return 1;
}

static int
mv88disable(SDev *sdev)
{
	char name[32];
	Ctlr *ctlr;

	ctlr = sdev->ctlr;
	snprint(name, sizeof name, "%s (%s)", sdev->name, sdev->ifc->name);
	intrdisable(ctlr->irq, mv88interrupt, ctlr, ctlr->tbdf, name);
	ctlr->enabled = 0;
	return 0;
}

static int
mv88verify(SDunit *unit)
{
	USED(unit);
	return 1;
}

static int
mv88online(SDunit *unit)
{
	Ctlr *ctlr;
	Reg *reg;

	ctlr = unit->dev->ctlr;
	reg = ctlr->mmio;

	if(reg->state1 & STonline != STonline)
		return 0;

	mv88cmd(ctlr, Creset, 0, nil);
	mv88cmd(ctlr, Ccond, 0, nil);
	return 1;
}

typedef struct Regs Regs;
struct Regs
{
	ulong offset;
	char *name;
};

static Regs regsctlr[] =
{
	0x00,	"sys addr low",
	0x02,	"sys addr high",
	0x04,	"block size",
	0x06,	"block count",
	0x08,	"argument low",
	0x0a,	"argument high",
	0x0c,	"transfer mode",
	0x0e,	"command",
	0x10,	"response 0",
	0x12,	"response 1",
	0x14,	"response 2",
	0x16,	"response 3",
	0x18,	"response 4",
	0x1a,	"response 5",
	0x1c,	"response 6",
	0x1e,	"response 7",
	0x20,	"buffer data port 0",
	0x22,	"buffer data port 1",
	0x24,	"present state 0",
	0x26,	"present state 1",
	0x28,	"host control",
	0x2a,	"block gap control",
	0x2c,	"clock control",
	0x2e,	"timeout/reset",
	0x30,	"normal interrupt status",
	0x32,	"error interrupt status",
	0x34,	"normal interrupt status enable",
	0x36,	"error interrupt status enable",
	0x38,	"normal interrupt status interrupt enable",
	0x3a,	"error interrupt status enable",
	0x3c,	"cmd12 error status",

	0x40,	"capabilities 0",
	0x42,	"capabilities 1",
	0x44,	"capabilities 2",
	0x46,	"capabilities 3",
	0x48,	"current 0",
	0x4a,	"current 1",
	0x4c,	"current 2",
	0x4e,	"current 3",

	0x60,	"io control",

	0x6a,	"command 1",
	0xFC,	"slot interrupt",
	0xFE,	"host control version",
};

static char*
rdregs(char *p, char *e, void *base, Regs *r, int n, char *prefix)
{
	int i;

	for(i = 0; i < n; i++)
		p = seprint(p, e, "%s%s%-19s %.4hux\n",
			prefix? prefix: "", prefix? ": ": "",
			r[i].name, *(ushort *)((uchar*)base + r[i].offset));
	return p;
}

static int
mv88rctl(SDunit *unit, char *p, int l)
{
	char *e, *op;
	Ctlr *ctlr;

	if((ctlr = unit->dev->ctlr) == nil)
		return 0;

	e = p+l;
	op = p;

	p = rdregs(p, e, ctlr->mmio, regsctlr, nelem(regsctlr), nil);
	return p-op;
}

SDifc sdmv88alp01ifc =  {
	"MV88ALP01 SD",

	mv88pnp,
	nil,	/* legacy */
	mv88enable,
	mv88disable,

	mv88verify,
	mv88online,
	nil /*mv88rio*/,
	mv88rctl,
	nil /*mv88wctl*/,

	nil /*scsibio*/,
	nil /*mv88probe*/,
	nil /*mv88clear*/,
	nil /*mv88rtopctl*/,
	nil /*mv88wtopctl*/,
};
